"""pylitterbot robot models."""

from datetime import datetime, timedelta, timezone

FEEDING_FILTER = f"""(
    order_by: {{ timestamp: desc }}
    where: {{
        _and: [
            {{ status: {{ _eq: dispensed }} }}
            {{ timestamp: {{ _gte: "{(datetime.now(timezone.utc) - timedelta(days=1)).isoformat()}" }} }}
        ]
    }}
)"""

FEEDER_ROBOT_MODEL = f"""
{{
    id
    name
    serial
    timezone
    isEighthCupEnabled
    created_at
    household_id
    state {{
        id
        info
        updated_at
        active_schedule {{
            id
            name
            meals
            created_at
        }}
    }}
    feeding_snack {FEEDING_FILTER} {{
        timestamp
        amount
    }}
    feeding_meal {FEEDING_FILTER} {{
        timestamp
        amount
        meal_name
        meal_number
        meal_total_portions
    }}
}}
"""

SLEEP_WAKE_ENABLED_MODEL = """
{
  sleepTime
  wakeTime
  isEnabled
}
"""

LITTER_ROBOT_4_MODEL = f"""
{{
    unitId
    name
    serial
    userId
    espFirmware
    picFirmwareVersion
    picFirmwareVersionHex
    laserBoardFirmwareVersion
    laserBoardFirmwareVersionHex
    wifiRssi
    unitPowerType
    catWeight
    displayCode
    unitTimezone
    unitTime
    cleanCycleWaitTime
    isKeypadLockout
    nightLightMode
    nightLightBrightness
    isPanelSleepMode
    panelSleepTime
    panelWakeTime
    weekdaySleepModeEnabled {{
        Sunday {SLEEP_WAKE_ENABLED_MODEL}
        Monday {SLEEP_WAKE_ENABLED_MODEL}
        Tuesday {SLEEP_WAKE_ENABLED_MODEL}
        Wednesday {SLEEP_WAKE_ENABLED_MODEL}
        Thursday {SLEEP_WAKE_ENABLED_MODEL}
        Friday {SLEEP_WAKE_ENABLED_MODEL}
        Saturday {SLEEP_WAKE_ENABLED_MODEL}
    }}
    unitPowerStatus
    sleepStatus
    robotStatus
    globeMotorFaultStatus
    pinchStatus
    catDetect
    isBonnetRemoved
    isNightLightLEDOn
    odometerPowerCycles
    odometerCleanCycles
    panelBrightnessHigh
    panelBrightnessLow
    smartWeightEnabled
    odometerEmptyCycles
    odometerFilterCycles
    isDFIResetPending
    DFINumberOfCycles
    DFILevelPercent
    isDFIFull
    DFIFullCounter
    DFITriggerCount
    litterLevel
    DFILevelMM
    isCatDetectPending
    globeMotorRetractFaultStatus
    robotCycleStatus
    robotCycleState
    weightSensor
    isOnline
    isOnboarded
    isProvisioned
    isDebugModeActive
    lastSeen
    sessionId
    setupDateTime
    isFirmwareUpdateTriggered
    firmwareUpdateStatus
    wifiModeStatus
    isUSBPowerOn
    USBFaultStatus
    isDFIPartialFull
    isLaserDirty
    surfaceType
    hopperStatus
    scoopsSavedCount
    isHopperRemoved
    optimalLitterLevel
    litterLevelPercentage
    litterLevelState
}}
"""



"""Litter-Robot 4."""

from __future__ import annotations

import logging
from datetime import datetime, time, timedelta, timezone
from enum import Enum, IntEnum, unique
from json import dumps
from typing import TYPE_CHECKING, Any, Dict, Union, cast
from uuid import uuid4

try:
    from zoneinfo import ZoneInfo
except ImportError:  # pragma: no cover
    from backports.zoneinfo import ZoneInfo  # type: ignore

from ..activity import Activity, Insight
from ..enums import LitterBoxStatus, LitterRobot4Command
from ..exceptions import InvalidCommandException, LitterRobotException
from ..utils import encode, to_enum, to_timestamp, utcnow
from .litterrobot import LitterRobot
from .models import LITTER_ROBOT_4_MODEL

if TYPE_CHECKING:
    from ..account import Account

_LOGGER = logging.getLogger(__name__)

LR4_ENDPOINT = "https://lr4.iothings.site/graphql"
LR4_STATUS_MAP = {
    "ROBOT_BONNET": LitterBoxStatus.BONNET_REMOVED,
    "ROBOT_CAT_DETECT": LitterBoxStatus.CAT_DETECTED,
    "ROBOT_CAT_DETECT_DELAY": LitterBoxStatus.CAT_SENSOR_TIMING,
    "ROBOT_CLEAN": LitterBoxStatus.CLEAN_CYCLE,
    "ROBOT_EMPTY": LitterBoxStatus.EMPTY_CYCLE,
    "ROBOT_FIND_DUMP": LitterBoxStatus.CLEAN_CYCLE,
    "ROBOT_IDLE": LitterBoxStatus.READY,
    "ROBOT_POWER_DOWN": LitterBoxStatus.POWER_DOWN,
    "ROBOT_POWER_OFF": LitterBoxStatus.OFF,
    "ROBOT_POWER_UP": LitterBoxStatus.POWER_UP,
}
ACTIVITY_STATUS_MAP: dict[str, LitterBoxStatus | str] = {
    "bonnetRemovedYes": LitterBoxStatus.BONNET_REMOVED,
    "catDetectStuckLaser": LitterBoxStatus.CAT_SENSOR_FAULT,
    "catWeight": "Pet Weight Recorded",
    "DFIFullFlagOn": LitterBoxStatus.DRAWER_FULL,
    "litterHopperDispensed": "Litter Dispensed",
    "odometerCleanCycles": "Clean Cycles",
    "powerTypeDC": "Battery Backup",
    "robotCycleStateCatDetect": LitterBoxStatus.CAT_SENSOR_INTERRUPTED,
    "robotCycleStatusDump": LitterBoxStatus.CLEAN_CYCLE,
    "robotCycleStatusIdle": LitterBoxStatus.CLEAN_CYCLE_COMPLETE,
    "robotStatusCatDetect": LitterBoxStatus.CAT_DETECTED,
}
CYCLE_STATE_STATUS_MAP = {
    "CYCLE_STATE_CAT_DETECT": LitterBoxStatus.CAT_SENSOR_INTERRUPTED,
    "CYCLE_STATE_PAUSE": LitterBoxStatus.PAUSED,
}
DISPLAY_CODE_STATUS_MAP = {"DC_CAT_DETECT": LitterBoxStatus.CAT_DETECTED}

LITTER_LEVEL_EMPTY = 500


@unique
class BrightnessLevel(IntEnum):
    """Brightness level of a Litter-Robot 4 unit."""

    LOW = 25
    MEDIUM = 50
    HIGH = 100


# Deprecated. Use BrightnessLevel.
NightLightLevel = BrightnessLevel


@unique
class FirmwareUpdateStatus(Enum):
    """Firmware update status."""

    NONE = "NONE"
    TRIGGERED = "TRIGGERED"
    PICTRIGGERED = "PICTRIGGERED"
    LASERBOARDTRIGGERED = "LASERBOARDTRIGGERED"
    ESPTRIGGERED = "ESPTRIGGERED"
    STARTED = "STARTED"
    IN_PROGRESS = "IN_PROGRESS"
    SUCCEEDED = "SUCCEEDED"
    FAILED = "FAILED"
    CANCELED = "CANCELED"
    DELETED = "DELETED"
    REJECTED = "REJECTED"
    TIMED_OUT = "TIMED_OUT"
    REMOVED = "REMOVED"
    COMPLETED = "COMPLETED"
    CANCELLATION_IN_PROGRESS = "CANCELLATION_IN_PROGRESS"
    DELETION_IN_PROGRESS = "DELETION_IN_PROGRESS"


@unique
class HopperStatus(Enum):
    """Hopper status."""

    ENABLED = "ENABLED"
    DISABLED = "DISABLED"
    MOTOR_FAULT_SHORT = "MOTOR_FAULT_SHORT"
    MOTOR_OT_AMPS = "MOTOR_OT_AMPS"
    MOTOR_DISCONNECTED = "MOTOR_DISCONNECTED"
    EMPTY = "EMPTY"


@unique
class LitterLevelState(Enum):
    """Litter level state."""

    OVERFILL = "OVERFILL"
    OPTIMAL = "OPTIMAL"
    REFILL = "REFILL"
    LOW = "LOW"
    EMPTY = "EMPTY"


@unique
class NightLightMode(Enum):
    """Night light mode of a Litter-Robot 4 unit."""

    OFF = "OFF"
    ON = "ON"
    AUTO = "AUTO"


@unique
class SurfaceType(Enum):
    """Surface type."""

    TILE = "TILE"
    CARPET = "CARPET"
    UNKNOWN = "UNKNOWN"


@unique
class UsbFaultStatus(Enum):
    """USB fault status."""

    NONE = "NONE"
    CLEAR = "CLEAR"
    SET = "SET"


@unique
class WifiModeStatus(Enum):
    """Wi-Fi mode status."""

    NONE = "NONE"
    OFF = "OFF"
    OFF_WAITING = "OFF_WAITING"
    OFF_CONNECTED = "OFF_CONNECTED"
    OFF_FAULT = "OFF_FAULT"
    HOTSPOT_WAITING = "HOTSPOT_WAITING"
    HOTSPOT_CONNECTED = "HOTSPOT_CONNECTED"
    HOTSPOT_FAULT = "HOTSPOT_FAULT"
    ROUTER_WAITING = "ROUTER_WAITING"
    ROUTER_CONNECTED = "ROUTER_CONNECTED"
    ROUTER_FAULT = "ROUTER_FAULT"


class LitterRobot4(LitterRobot):  # pylint: disable=abstract-method
    """Data and methods for interacting with a Litter-Robot 4 automatic, self-cleaning litter box."""

    _attr_model = "Litter-Robot 4"

    VALID_WAIT_TIMES = [3, 7, 8, 15, 25, 30]

    _data_cycle_capacity = "DFINumberOfCycles"
    _data_cycle_count = "odometerCleanCycles"
    _data_drawer_full_cycles = "DFIFullCounter"
    _data_id = "unitId"
    _data_name = "name"
    _data_power_status = "unitPowerType"
    _data_serial = "serial"
    _data_setup_date = "setupDateTime"

    _command_clean = LitterRobot4Command.CLEAN_CYCLE
    _command_night_light_off = LitterRobot4Command.NIGHT_LIGHT_MODE_OFF
    _command_night_light_on = LitterRobot4Command.NIGHT_LIGHT_MODE_AUTO
    _command_panel_lock_off = LitterRobot4Command.KEY_PAD_LOCK_OUT_OFF
    _command_panel_lock_on = LitterRobot4Command.KEY_PAD_LOCK_OUT_ON
    _command_power_off = LitterRobot4Command.POWER_OFF
    _command_power_on = LitterRobot4Command.POWER_ON

    _litter_level = LITTER_LEVEL_EMPTY
    _litter_level_exp: datetime | None = None

    _firmware_details: dict[str, bool | dict[str, str]] | None = None
    _firmware_details_requested: datetime | None = None

    def __init__(self, data: dict, account: Account) -> None:
        """Initialize a Litter-Robot 4."""
        super().__init__(data, account)
        self._path = LR4_ENDPOINT

    @property
    def clean_cycle_wait_time_minutes(self) -> int:
        """Return the number of minutes after a cat uses the Litter-Robot to begin an automatic clean cycle."""
        return cast(int, self._data.get("cleanCycleWaitTime", 7))

    @property
    def firmware(self) -> str:
        """Return the firmware version."""
        return (
            f"ESP: {self._data.get('espFirmware')} / "
            f"PIC: {self._data.get('picFirmwareVersion')} / "
            f"TOF: {self._data.get('laserBoardFirmwareVersion')}"
        )

    @property
    def firmware_update_status(self) -> str:
        """Return the firmware update status."""
        return cast(str, self._data.get("firmwareUpdateStatus", "UNKNOWN"))

    @property
    def firmware_update_triggered(self) -> bool:
        """Return `True` if a firmware update has been triggered."""
        return self._data.get("isFirmwareUpdateTriggered") is True

    @property
    def hopper_status(self) -> HopperStatus | None:
        """Return the hopper status."""
        return to_enum(self._data.get("hopperStatus"), HopperStatus)

    @property
    def is_drawer_full_indicator_triggered(self) -> bool:
        """Return `True` if the drawer full indicator has been triggered."""
        return self._data.get("isDFIFull") is True

    @property
    def is_hopper_removed(self) -> bool | None:
        """Return `True` if the hopper is removed/disabled."""
        return self._data.get("isHopperRemoved") is True

    @property
    def is_online(self) -> bool:
        """Return `True` if the robot is online."""
        return self._data.get("isOnline") is True

    @property
    def is_sleeping(self) -> bool:
        """Return `True` if the Litter-Robot is currently "sleeping" and won't automatically perform a clean cycle."""
        return bool(self._data.get("sleepStatus", "WAKE") != "WAKE")

    @property
    def is_waste_drawer_full(self) -> bool:
        """Return `True` if the Litter-Robot is reporting that the waste drawer is full."""
        return self._data.get("isDFIFull") is True

    @property
    def litter_level(self) -> float:
        """Return the litter level."""
        return cast(float, self._data.get("litterLevelPercentage", 0)) * 100

    @property
    def litter_level_calculated(self) -> float:
        """Return the calculated litter level.

        The litterLevel field from the API is a millimeter distance to the
        top center time of flight (ToF) sensor and is interpreted as:

        ~ 441 full
        ~ 451 nominal
        ~ 461 low
        ~ 471 very low
        """
        new_level = int(self._data.get("litterLevel", LITTER_LEVEL_EMPTY))
        now = datetime.now(timezone.utc)
        if self._data.get("robotStatus") == "ROBOT_CLEAN":
            self._litter_level_exp = now + timedelta(minutes=1)
        elif (
            self._litter_level_exp is None
            or self._litter_level_exp < now
            or abs(self._litter_level - new_level) < 10
        ):
            self._litter_level = new_level
        return max(round(100 - (self._litter_level - 440) / 0.6, -1), 0)

    @property
    def litter_level_state(self) -> LitterLevelState | None:
        """Return the litter level state."""
        return to_enum(self._data.get("litterLevelState"), LitterLevelState)

    @property
    def night_light_brightness(self) -> int:
        """Return the night light brightness."""
        return int(self._data.get("nightLightBrightness", 0))

    @property
    def night_light_level(self) -> BrightnessLevel | None:
        """Return the night light level."""
        return to_enum(self.night_light_brightness, BrightnessLevel, False)

    @property
    def night_light_mode(self) -> NightLightMode | None:
        """Return the night light mode setting."""
        return to_enum(self._data.get("nightLightMode"), NightLightMode)

    @property
    def night_light_mode_enabled(self) -> bool:
        """Return `True` if night light mode is enabled."""
        return bool(self._data.get("nightLightMode", "OFF") != "OFF")

    @property
    def panel_brightness(self) -> BrightnessLevel | None:
        """Return the panel brightness."""
        brightness = self._data.get("panelBrightnessHigh")
        if brightness in list(BrightnessLevel):
            return BrightnessLevel(brightness)
        return None

    @property
    def panel_lock_enabled(self) -> bool:
        """Return `True` if the buttons on the robot are disabled."""
        return self._data.get("isKeypadLockout") is True

    @property
    def pet_weight(self) -> float:
        """Return the last recorded pet weight in pounds (lbs)."""
        return cast(float, self._data.get("catWeight", 0))

    @property
    def scoops_saved_count(self) -> int:
        """Return the scoops saved count."""
        return cast(int, self._data.get("scoopsSavedCount", 0))

    @property
    def sleep_mode_enabled(self) -> bool:
        """Return `True` if sleep mode is enabled."""
        sleep_schedule = self._data["weekdaySleepModeEnabled"]
        return any(day["isEnabled"] for day in sleep_schedule.values())

    @property
    def sleep_mode_start_time(self) -> datetime | None:
        """Return the sleep mode start time, if any."""
        self._revalidate_sleep_info()
        return self._sleep_mode_start_time

    @property
    def sleep_mode_end_time(self) -> datetime | None:
        """Return the sleep mode end time, if any."""
        self._revalidate_sleep_info()
        return self._sleep_mode_end_time

    @property
    def status(self) -> LitterBoxStatus:
        """Return the status of the Litter-Robot.

        The Litter-Robot 4's status is determined based on the values of:
          - `displayCode`
          - `robotCycleState`
          - `robotStatus`
          - `isDFIFull`
          - `isOnline`
        """
        if not self.is_online:
            return LitterBoxStatus.OFFLINE
        if status := CYCLE_STATE_STATUS_MAP.get(self._data["robotCycleState"]):
            return status
        status = LR4_STATUS_MAP.get(self._data["robotStatus"], LitterBoxStatus.UNKNOWN)
        if status == LitterBoxStatus.READY:
            if display_code := DISPLAY_CODE_STATUS_MAP.get(self._data["displayCode"]):
                return display_code
            if self.is_waste_drawer_full:
                return LitterBoxStatus.DRAWER_FULL
        return status

    @property
    def status_code(self) -> str | None:
        """Return the status code of the Litter-Robot."""
        return (
            self.status.value
            if self.status != LitterBoxStatus.UNKNOWN
            else self._data.get("robotStatus")
        )

    @property
    def surface_type(self) -> SurfaceType | None:
        """Return the surface type."""
        return to_enum(self._data.get("surfaceType"), SurfaceType)

    @property
    def usb_fault_status(self) -> UsbFaultStatus | None:
        """Return the USB fault status."""
        return to_enum(self._data.get("USBFaultStatus"), UsbFaultStatus)

    @property
    def waste_drawer_level(self) -> float:
        """Return the approximate waste drawer level."""
        return cast(float, self._data.get("DFILevelPercent", 0))

    @property
    def wifi_mode_status(self) -> WifiModeStatus | None:
        """Return the Wi-Fi mode status."""
        return to_enum(self._data.get("wifiModeStatus"), WifiModeStatus)

    def _revalidate_sleep_info(self) -> None:
        """Revalidate sleep info."""
        if (
            self.sleep_mode_enabled
            and (now := utcnow()) > (self._sleep_mode_start_time or now)
            and now > (self._sleep_mode_end_time or now)
        ):
            self._parse_sleep_info()

    def _parse_activity(self, activity: dict[str, str]) -> LitterBoxStatus | str:
        """Parse an activity."""
        value = activity["value"]
        action = ACTIVITY_STATUS_MAP.get(value, value)
        if value == "catWeight":
            action = f"{action}: {activity['actionValue']} lbs"
        if value == self._data_cycle_count:
            action = f"{action}: {activity['actionValue']}"
        if value == "litterHopperDispensed":
            # TODO: figure out what this value refers to. Could it be a fraction of lbs? Number of motor revolutions?
            action = f"{action}: {activity['actionValue']}"
        return action

    def _parse_sleep_info(self) -> None:
        """Parse the sleep info of a Litter-Robot."""
        start = end = None
        now = datetime.now(ZoneInfo(self._data["unitTimezone"]))
        sleep_schedule = self._data["weekdaySleepModeEnabled"]
        for idx in range(-7, 8):
            day = now + timedelta(days=idx)
            if (schedule := sleep_schedule[day.strftime("%A")])["isEnabled"]:
                start_of_day = datetime.combine(day, time(), day.tzinfo)
                if (wake_time := schedule["wakeTime"]) < (
                    sleep_time := schedule["sleepTime"]
                ):
                    start = start_of_day - timedelta(minutes=1440 - sleep_time)
                else:
                    start = start_of_day + timedelta(minutes=sleep_time)
                if now > start or end is None:
                    end = start_of_day + timedelta(minutes=wake_time)
                if now > max(start, end):
                    continue
                break
        self._sleep_mode_start_time = start
        self._sleep_mode_end_time = end

    async def _dispatch_command(self, command: str, **kwargs: Any) -> bool:
        """Send a command to the Litter-Robot."""
        try:
            data = await self._post(
                json={
                    "query": """
                        mutation sendCommand(
                            $serial: String!
                            $command: String!
                            $value: String
                            $commandSource: String
                        ) {
                        sendLitterRobot4Command(
                            input: {
                            serial: $serial
                            command: $command
                            value: $value
                            commandSource: $commandSource
                            }
                        )
                    }
                    """,
                    "variables": {"serial": self.serial, "command": command, **kwargs},
                }
            )
            data = cast(dict, data)
            error = data.get("data", {}).get("sendLitterRobot4Command", "")
            if error and "Error" in error:
                raise InvalidCommandException(error)
            if error := ", ".join(e.get("message", "") for e in data.get("errors", {})):
                raise InvalidCommandException(error)
            return True
        except InvalidCommandException as ex:
            _LOGGER.error(ex)
            return False

    async def refresh(self) -> None:
        """Refresh the Litter-Robot's data from the API."""
        data = await self._post(
            json={
                "query": f"""
                    query GetLR4($serial: String!) {{
                        getLitterRobot4BySerial(serial: $serial) {LITTER_ROBOT_4_MODEL}
                    }}
                    """,
                "variables": {"serial": self.serial},
            },
        )
        data = cast(dict, data)
        self._update_data(data.get("data", {}).get("getLitterRobot4BySerial", {}))

    async def reset(self) -> bool:
        """Perform a reset on the Litter-Robot.

        Clears errors and may trigger a cycle. Make sure the globe is clear before proceeding.
        """
        return await self._dispatch_command(LitterRobot4Command.SHORT_RESET_PRESS)

    async def set_name(self, name: str) -> bool:
        """Set the name."""
        data = await self._post(
            json={
                "query": """
                    mutation rename(
                        $serial: String!
                        $name: String
                    ) {
                        updateLitterRobot4(
                            input: {
                                serial: $serial
                                name: $name
                            }
                        ) {
                            name
                        }
                    }
                """,
                "variables": {"serial": self.serial, "name": name},
            }
        )
        updated_data = cast(dict, data).get("data", {}).get("updateLitterRobot4", {})
        self._update_data(updated_data, partial=True)
        return self.name == name

    async def set_night_light_brightness(
        self, brightness: int | BrightnessLevel
    ) -> bool:
        """Set the night light brightness on the robot."""
        if brightness not in list(BrightnessLevel):
            raise InvalidCommandException(
                f"Attempt to send an invalid night light level to Litter-Robot. "
                f"Brightness must be one of: {list(BrightnessLevel)}, but received {brightness}"
            )
        return await self._dispatch_command(
            LitterRobot4Command.SET_NIGHT_LIGHT_VALUE,
            value=dumps({"nightLightPower": int(brightness)}),
        )

    async def set_night_light_mode(self, mode: NightLightMode) -> bool:
        """Set the night light mode on the robot."""
        mode_to_command = {
            NightLightMode.ON: LitterRobot4Command.NIGHT_LIGHT_MODE_ON,
            NightLightMode.OFF: LitterRobot4Command.NIGHT_LIGHT_MODE_OFF,
            NightLightMode.AUTO: LitterRobot4Command.NIGHT_LIGHT_MODE_AUTO,
        }
        return await self._dispatch_command(mode_to_command[mode])

    async def set_panel_brightness(self, brightness: BrightnessLevel) -> bool:
        """Set the panel brightness."""
        level_to_command = {
            BrightnessLevel.LOW: LitterRobot4Command.PANEL_BRIGHTNESS_LOW,
            BrightnessLevel.MEDIUM: LitterRobot4Command.PANEL_BRIGHTNESS_MEDIUM,
            BrightnessLevel.HIGH: LitterRobot4Command.PANEL_BRIGHTNESS_HIGH,
        }
        return await self._dispatch_command(level_to_command[brightness])

    async def set_wait_time(self, wait_time: int) -> bool:
        """Set the wait time on the Litter-Robot."""
        if wait_time not in self.VALID_WAIT_TIMES:
            raise InvalidCommandException(
                f"Attempt to send an invalid wait time to Litter-Robot. Wait time must be one of: {self.VALID_WAIT_TIMES}, but received {wait_time}"
            )
        return await self._dispatch_command(
            LitterRobot4Command.SET_CLUMP_TIME,
            value=dumps({"clumpTime": wait_time}),
        )

    async def get_activity_history(self, limit: int = 100) -> list[Activity]:
        """Return the activity history."""
        if limit < 1:
            raise InvalidCommandException(
                f"Invalid range for parameter limit, value: {limit}, valid range: 1-inf"
            )
        data = await self._post(
            json={
                "query": """
                    query GetLR4Activity(
                        $serial: String!
                        $startTimestamp: String
                        $endTimestamp: String
                        $limit: Int
                        $consumer: String
                    ) {
                        getLitterRobot4Activity(
                            serial: $serial
                            startTimestamp: $startTimestamp
                            endTimestamp: $endTimestamp
                            limit: $limit
                            consumer: $consumer
                        ) {
                            serial
                            measure
                            timestamp
                            value
                            actionValue
                            originalHex
                            valueString
                            stateString
                            consumer
                            commandSource
                        }
                    }
                """,
                "variables": {
                    "serial": self.serial,
                    "limit": limit,
                    "consumer": "app",
                },
            }
        )
        activities = cast(dict, data).get("data", {}).get("getLitterRobot4Activity", {})
        if activities is None:
            raise LitterRobotException("Activity history could not be retrieved.")
        return [
            Activity(timestamp, self._parse_activity(activity))
            for activity in activities
            if (timestamp := to_timestamp(activity["timestamp"])) is not None
        ]

    async def get_insight(
        self, days: int = 30, timezone_offset: int | None = None
    ) -> Insight:
        """Return the insight data."""
        data = await self._post(
            json={
                "query": """
                    query GetLR4Insights(
                        $serial: String!
                        $startTimestamp: String
                        $timezoneOffset: Int
                    ) {
                        getLitterRobot4Insights(
                            serial: $serial
                            startTimestamp: $startTimestamp
                            timezoneOffset: $timezoneOffset
                        ) {
                            totalCycles
                            averageCycles
                            cycleHistory {
                                date
                                numberOfCycles
                            }
                            totalCatDetections
                        }
                    }
                """,
                "variables": {
                    "serial": self.serial,
                    "startTimestamp": (utcnow() - timedelta(days=days)).strftime(
                        "%Y-%m-%dT%H:%M:%S.%fZ"
                    ),
                    "timezoneOffset": timezone_offset,
                },
            }
        )
        insight = cast(dict, data).get("data", {}).get("getLitterRobot4Insights", {})
        if insight is None:
            raise LitterRobotException("Insight data could not be retrieved.")
        return Insight(
            insight["totalCycles"],
            insight["averageCycles"],
            [
                (
                    datetime.strptime(cycle["date"], "%Y-%m-%d").date(),
                    cycle["numberOfCycles"],
                )
                for cycle in insight["cycleHistory"]
            ],
        )

    async def get_firmware_details(
        self, force_check: bool = False
    ) -> dict[str, bool | dict[str, str]] | None:
        """Get the firmware details."""
        if (
            force_check
            or not self._firmware_details
            or (requested := self._firmware_details_requested) is None
            or requested + timedelta(minutes=15) < utcnow()
        ):
            data = await self._post(
                json={
                    "query": """
                        query getFirmwareDetails($serial: String!) {
                            litterRobot4CompareFirmwareVersion(serial: $serial) {
                                isEspFirmwareUpdateNeeded
                                isPicFirmwareUpdateNeeded
                                isLaserboardFirmwareUpdateNeeded
                                latestFirmware {
                                    espFirmwareVersion
                                    picFirmwareVersion
                                    laserBoardFirmwareVersion
                                }
                            }
                        }
                    """,
                    "variables": {"serial": self.serial},
                }
            )
            self._firmware_details = (
                cast(Dict[str, Dict[str, Dict[str, Union[bool, Dict[str, str]]]]], data)
                .get("data", {})
                .get("litterRobot4CompareFirmwareVersion", {})
            )
            self._firmware_details_requested = utcnow()
        return self._firmware_details

    async def get_latest_firmware(self, force_check: bool = False) -> str | None:
        """Get the latest firmware available."""
        if (firmware := await self.get_firmware_details(force_check)) is None:
            return None

        latest_firmware = cast(Dict[str, str], firmware.get("latestFirmware", {}))
        return (
            f"ESP: {latest_firmware.get('espFirmwareVersion')} / "
            f"PIC: {latest_firmware.get('picFirmwareVersion')} / "
            f"TOF: {latest_firmware.get('laserBoardFirmwareVersion')}"
        )

    async def has_firmware_update(self, force_check: bool = False) -> bool:
        """Check if a firmware update is available."""
        if (firmware := await self.get_firmware_details(force_check)) is None:
            return False
        return any(value for value in firmware.values() if isinstance(value, bool))

    async def update_firmware(self) -> bool:
        """Trigger a firmware update."""
        data = await self._post(
            json={
                "query": """
                    mutation updateFirmware($serial: String!) {
                        litterRobot4TriggerFirmwareUpdate(input: {serial: $serial}) {
                            isUpdateTriggered
                            isEspFirmwareUpdateNeeded
                            isPicFirmwareUpdateNeeded
                            isLaserboardFirmwareUpdateNeeded
                        }
                    }
                """,
                "variables": {"serial": self.serial},
            }
        )
        data = cast(dict, data)
        firmware = data.get("data", {}).get("litterRobot4TriggerFirmwareUpdate", {})
        return bool(firmware.get("isUpdateTriggered", False))

    async def toggle_hopper(self, is_removed: bool) -> bool:
        """Enables/Disables the LitterHopper. A disabled hopper is synonymous with being removed. Returns `True` if request was successful.

        NOTE: The API appears to take ~5 seconds to reflect the data changes made by this method. To prevent surprises, this method will fake the expected changes to Hopper-related data properties before returning. If you need a guaranteed state, refresh the data afterwards.
        """
        data = await self._post(
            json={
                "query": """
                    mutation ToggleHopper($serial: String!, $isRemoved: Boolean!) {
                        toggleHopper(serial: $serial, isRemoved: $isRemoved) {
                            success
                        }
                    }
                """,
                "variables": {"serial": self.serial, "isRemoved": is_removed},
            }
        )
        data = cast(dict, data)
        toggle_hopper = data.get("data", {}).get("toggleHopper", {})
        is_success = bool(toggle_hopper.get("success", False))
        if is_success:
            # data is now stale, hopper-related properties will have changed as
            # a consequence. This mutation doesn't expose robot data for a
            # partial update, so we splice in reasonable data changes instead.
            #
            # Unfortunately `hopperStatus` doesn't instantaneously update to
            # reflect the changes from the mutation, so we can't perform a
            # simple data refresh. This is likely due to the LitterHopper
            # device reporting back its own status back to the API.
            self._update_data(
                {
                    "isHopperRemoved": is_removed,
                    "hopperStatus": "DISABLED" if is_removed else "ENABLED",
                },
                partial=True,
            )
        return is_success

    async def send_subscribe_request(self, send_stop: bool = False) -> None:
        """Send a subscribe request and, optionally, unsubscribe from a previous subscription."""
        if not self._ws:
            return
        if send_stop:
            await self.send_unsubscribe_request()
        self._ws_subscription_id = str(uuid4())

        await self._ws.send_json(
            {
                "id": self._ws_subscription_id,
                "payload": {
                    "data": dumps(
                        {
                            "query": f"""
                                subscription GetLR4($serial: String!) {{
                                    litterRobot4StateSubscriptionBySerial(serial: $serial) {LITTER_ROBOT_4_MODEL}
                                }}
                            """,
                            "variables": {"serial": self.serial},
                        }
                    ),
                    "extensions": {
                        "authorization": {
                            "Authorization": await self._account.get_bearer_authorization(),
                            "host": "lr4.iothings.site",
                        }
                    },
                },
                "type": "start",
            }
        )

    @staticmethod
    async def get_websocket_config(account: Account) -> dict[str, Any]:
        """Get wesocket config."""
        return {
            "url": f"{LR4_ENDPOINT}/realtime",
            "params": {
                "header": encode(
                    {
                        "Authorization": await account.get_bearer_authorization(),
                        "host": "lr4.iothings.site",
                    }
                ),
                "payload": encode({}),
            },
            "headers": {"sec-websocket-protocol": "graphql-ws"},
        }

    @staticmethod
    def parse_websocket_message(data: dict) -> dict | None:
        """Parse a wesocket message."""
        if (data_type := data["type"]) == "data":
            data = data["payload"]["data"]["litterRobot4StateSubscriptionBySerial"]
            return data
        if data_type == "error":
            _LOGGER.error(data)
        elif data_type not in ("start_ack", "ka", "complete"):
            _LOGGER.debug(data)
        return None



"""Websocket monitor."""

from __future__ import annotations

import asyncio
import logging
from datetime import datetime
from json import loads
from random import uniform
from typing import TYPE_CHECKING

from aiohttp import ClientWebSocketResponse, WSMsgType

from .robot import Robot
from .utils import utcnow

if TYPE_CHECKING:
    from .account import Account

_LOGGER = logging.getLogger(__name__)


async def cancel_task(*tasks: asyncio.Task | None) -> None:
    """Cancel task(s)."""
    for task in tasks:
        if task is not None and not task.done():
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass


class WebSocketMonitor:
    """Web socket monitor for a robot."""

    def __init__(self, account: Account, robot_class: type[Robot]) -> None:
        """Initialize a web socket monitor."""
        self._account = account
        self._robot_class = robot_class
        self._disconnect = False
        self._ws: ClientWebSocketResponse | None = None
        self._monitor_task: asyncio.Task | None = None
        self._receiver_task: asyncio.Task | None = None
        self._last_received: datetime | None = None

    @property
    def connected(self) -> bool:
        """Return `True` if the web socket is connected."""
        if self._disconnect:
            return False
        return False if self._ws is None else not self._ws.closed

    @property
    def websocket(self) -> ClientWebSocketResponse | None:
        """Return the web socket."""
        return self._ws

    @property
    def monitor(self) -> asyncio.Task | None:
        """Return the monitor task."""
        return self._monitor_task

    async def new_connection(self, start_monitor: bool = False) -> None:
        """Create a new connection and, optionally, start the monitor."""
        await cancel_task(self._receiver_task)
        self._disconnect = False
        self._ws = await self._account.session.websession.ws_connect(
            **await self._robot_class.get_websocket_config(self._account)
        )
        self._receiver_task = asyncio.ensure_future(self._receiver())
        if start_monitor:
            await self.start_monitor()

    async def _receiver(self) -> None:
        """Receive a message from a web socket."""
        if not (websocket := self._ws):
            return
        while not websocket.closed:
            try:
                msg = await websocket.receive(timeout=300)
                if msg.type in (WSMsgType.CLOSE, WSMsgType.CLOSING, WSMsgType.CLOSED):
                    break
                self._last_received = utcnow()
                if msg.type == WSMsgType.TEXT:
                    m_data = loads(msg.data)
                    # pylint: disable=protected-access
                    if (data := self._robot_class.parse_websocket_message(m_data)) and (
                        robot := self._account.get_robot(
                            data.get(self._robot_class._data_id)
                        )
                    ):
                        robot._update_data(data)
                elif msg.type == WSMsgType.ERROR:
                    self._log_message(msg, True)
                    continue
            except asyncio.TimeoutError:
                for robot in self._account.get_robots(self._robot_class):
                    await robot.send_subscribe_request(send_stop=True)
        self._log_message("web socket stopped")

    async def _monitor(self) -> None:
        """Monitor a web socket connection."""
        attempt = 0
        while not self._disconnect:
            while self.connected:
                await asyncio.sleep(1)
            if not self._disconnect:
                try:
                    await self.new_connection()
                except Exception as ex:  # pylint: disable=broad-except
                    self._log_message(ex, True)
                if not self._ws or self._ws.closed:
                    await asyncio.sleep(min(1 * 2**attempt + uniform(0, 1), 300))
                    attempt += 1
                    continue
                attempt = 0
                self._log_message("web socket connection reopened")
                for robot in self._account.get_robots(self._robot_class):
                    await robot.subscribe()

    async def start_monitor(self) -> None:
        """Start or restart the monitor task."""
        if self._monitor_task is None or self._monitor_task.done():
            self._monitor_task = asyncio.ensure_future(self._monitor())

    async def stop_monitor(self) -> None:
        """Stop the monitor task."""
        await cancel_task(self._monitor_task)

    async def close(self) -> None:
        """Close the web socket."""
        self._disconnect = True
        if self._ws:
            await self._ws.close()
        await cancel_task(self._monitor_task, self._receiver_task)

    def _log_message(self, message: str | Exception, is_error: bool = False) -> None:
        """Log a message."""
        log_method = _LOGGER.error if is_error else _LOGGER.debug
        log_method("%s %s", self._robot_class.__name__, message)



"""Utilities module."""

from __future__ import annotations

import json
import logging
import re
from base64 import b64decode, b64encode
from collections.abc import Iterable, Mapping
from datetime import datetime, time, timezone
from enum import Enum
from typing import Any, Type, TypeVar, cast, overload
from urllib.parse import urljoin as _urljoin
from warnings import warn

_LOGGER = logging.getLogger(__name__)
_T = TypeVar("_T")
_E = TypeVar("_E", bound=Enum)

ENCODING = "utf-8"
REDACTED = "**REDACTED**"
REDACT_FIELDS = [
    "token",
    "access_token",
    "id_token",
    "idToken",
    "refresh_token",
    "refreshToken",
    "userId",
    "userEmail",
    "sessionId",
    "oneSignalPlayerId",
    "deviceId",
    "id",
    "litterRobotId",
    "unitId",
    "litterRobotSerial",
    "serial",
    "s3ImageURL",
]


def decode(value: str) -> str:
    """Decode a value."""
    return b64decode(value).decode(ENCODING)


def encode(value: str | dict) -> str:
    """Encode a value."""
    if isinstance(value, dict):
        value = json.dumps(value)
    return b64encode(value.encode(ENCODING)).decode(ENCODING)


def to_timestamp(timestamp: str | None) -> datetime | None:
    """Construct a UTC offset-aware datetime from a Litter-Robot API timestamp."""
    if not timestamp:
        return None
    if "Z" in timestamp:
        timestamp = timestamp.replace("Z", "")
    if (utc_offset := "+00:00") not in timestamp:
        timestamp += utc_offset
    timestamp = re.sub(r"(\.\d+)", lambda m: m.group().ljust(7, "0")[:7], timestamp)
    return datetime.fromisoformat(timestamp)


def pluralize(word: str, count: int) -> str:
    """Pluralize a word."""
    return f"{count} {word}{'s' if count != 1 else ''}"


def round_time(_datetime: datetime | None = None, round_to: int = 60) -> datetime:
    """Round a datetime to the specified seconds or 1 minute if not specified."""
    if not _datetime:
        _datetime = utcnow()

    return datetime.fromtimestamp(
        (_datetime.timestamp() + round_to / 2) // round_to * round_to, _datetime.tzinfo
    )


def today_at_time(_time: time) -> datetime:
    """Return a datetime representing today at the passed in time."""
    return datetime.combine(utcnow().astimezone(_time.tzinfo), _time)


def urljoin(base: str, subpath_or_url: str | None) -> str:
    """Join a base URL and subpath or URL to form an absolute interpretation of the latter."""
    if not subpath_or_url:
        return base
    if not base.endswith("/"):
        base += "/"
    return _urljoin(base, subpath_or_url)


def utcnow() -> datetime:
    """Return the current UTC offset-aware datetime."""
    return datetime.now(timezone.utc)


def send_deprecation_warning(
    old_name: str, new_name: str | None = None
) -> None:  # pragma: no cover
    """Log a deprecation warning message."""
    message = f"{old_name} has been deprecated{'' if new_name is None else f' in favor of {new_name}'} and will be removed in a future release"
    warn(message, DeprecationWarning, stacklevel=2)
    _LOGGER.warning(message)


@overload
def redact(data: Mapping) -> dict: ...


@overload
def redact(data: _T) -> _T: ...


def redact(data: _T) -> _T:
    """Redact sensitive data in a dict."""
    if not isinstance(data, (Mapping, list)):
        return data

    if isinstance(data, list):
        return cast(_T, [redact(val) for val in data])

    redacted = {**data}

    for key, value in redacted.items():
        if value is None:
            continue
        if isinstance(value, str) and not value:
            continue
        if key in REDACT_FIELDS:
            redacted[key] = REDACTED
        elif isinstance(value, Mapping):
            redacted[key] = redact(value)
        elif isinstance(value, list):
            redacted[key] = [redact(item) for item in value]

    return cast(_T, redacted)


def first_value(
    data: dict | None,
    keys: Iterable,
    default: Any | None = None,
    return_none: bool = False,
) -> Any | None:
    """Return the first valid key's value."""
    if not data:
        return default
    for key in keys:
        if key in data and ((value := data[key]) is not None or return_none):
            return value
    return default


def to_enum(value: Any, typ: Type[_E], log_warning: bool = True) -> _E | None:
    """Get the corresponding enum member from a value."""
    if value is None:
        return None
    try:
        return typ(value)
    except ValueError:
        if log_warning:
            logging.warning("Value '%s' not found in enum %s", value, typ.__name__)
    except (AttributeError, TypeError):
        logging.error("Provided class %s is not a valid Enum", typ)
    return None



"""pylitterbot enums."""

from __future__ import annotations

import logging
from enum import Enum
from typing import Any

_LOGGER = logging.getLogger(__name__)


class FeederRobotCommand:
    """Known commands that can be sent to trigger an action or setting for a Feeder-Robot."""

    GIVE_SNACK = "giveSnack"
    SET_AUTO_NIGHT_MODE = "setAutoNightMode"
    SET_GRAVITY_MODE = "setGravityMode"
    SET_PANEL_LOCKOUT = "setPanelLockout"


class LitterBoxCommand:
    """Known commands that can be sent to trigger an action or setting for Litter-Robot 3."""

    ENDPOINT = "dispatch-commands"
    PREFIX = "<"

    CLEAN = "C"
    DEFAULT_SETTINGS = "D"
    LOCK_OFF = "L0"
    LOCK_ON = "L1"
    NIGHT_LIGHT_OFF = "N0"
    NIGHT_LIGHT_ON = "N1"
    POWER_OFF = "P0"
    POWER_ON = "P1"
    # REFRESH = "R"  # valid command, not sure what it does though, reset or refresh maybe
    SLEEP_MODE_OFF = "S0"
    SLEEP_MODE_ON = "S1"
    WAIT_TIME = "W"


class LitterRobot4Command:
    """Known commands that can be sent to trigger an action or setting for a Litter-Robot 4."""

    CLEAN_CYCLE = "cleanCycle"
    KEY_PAD_LOCK_OUT_OFF = "keyPadLockOutOff"
    KEY_PAD_LOCK_OUT_ON = "keyPadLockOutOn"
    NIGHT_LIGHT_MODE_AUTO = "nightLightModeAuto"
    NIGHT_LIGHT_MODE_OFF = "nightLightModeOff"
    NIGHT_LIGHT_MODE_ON = "nightLightModeOn"
    PANEL_BRIGHTNESS_LOW = "panelBrightnessLow"
    PANEL_BRIGHTNESS_MEDIUM = "panelBrightnessMed"
    PANEL_BRIGHTNESS_HIGH = "panelBrightnessHigh"
    POWER_OFF = "powerOff"
    POWER_ON = "powerOn"
    REQUEST_STATE = "requestState"
    SET_CLUMP_TIME = "setClumpTime"
    SET_NIGHT_LIGHT_VALUE = "setNightLightValue"
    SHORT_RESET_PRESS = "shortResetPress"


class LitterBoxStatusMixIn:
    """Litter box status mixin."""

    _text: str | None
    _minimum_cycles_left: int


class LitterBoxStatus(LitterBoxStatusMixIn, Enum):
    """Representation of a Litter-Robot status."""

    def __new__(
        cls, value: str | None, text: str | None = None, minimum_cycles_left: int = 3
    ) -> LitterBoxStatus:
        """Create and return a new Litter Box Status."""
        obj = object.__new__(cls)
        obj._value_ = value
        obj._text = text
        obj._minimum_cycles_left = minimum_cycles_left
        return obj

    BONNET_REMOVED = ("BR", "Bonnet Removed")
    CLEAN_CYCLE_COMPLETE = ("CCC", "Clean Cycle Complete")
    CLEAN_CYCLE = ("CCP", "Clean Cycle In Progress")
    CAT_DETECTED = ("CD", "Cat Detected")
    CAT_SENSOR_FAULT = ("CSF", "Cat Sensor Fault")
    CAT_SENSOR_INTERRUPTED = ("CSI", "Cat Sensor Interrupted")
    CAT_SENSOR_TIMING = ("CST", "Cat Sensor Timing")
    DRAWER_FULL_1 = ("DF1", "Drawer Almost Full - 2 Cycles Left", 2)
    DRAWER_FULL_2 = ("DF2", "Drawer Almost Full - 1 Cycle Left", 1)
    DRAWER_FULL = ("DFS", "Drawer Full", 0)
    DUMP_HOME_POSITION_FAULT = ("DHF", "Dump + Home Position Fault")
    DUMP_POSITION_FAULT = ("DPF", "Dump Position Fault")
    EMPTY_CYCLE = ("EC", "Empty Cycle")
    HOME_POSITION_FAULT = ("HPF", "Home Position Fault")
    OFF = ("OFF", "Off")
    OFFLINE = ("OFFLINE", "Offline")
    OVER_TORQUE_FAULT = ("OTF", "Over Torque Fault")
    PAUSED = ("P", "Clean Cycle Paused")
    PINCH_DETECT = ("PD", "Pinch Detect")
    POWER_DOWN = ("PWRD", "Powering Down")
    POWER_UP = ("PWRU", "Powering Up")
    READY = ("RDY", "Ready")
    STARTUP_CAT_SENSOR_FAULT = ("SCF", "Cat Sensor Fault At Startup")
    STARTUP_DRAWER_FULL = ("SDF", "Drawer Full At Startup", 0)
    STARTUP_PINCH_DETECT = ("SPF", "Pinch Detect At Startup")

    # Handle unknown/future unit statuses
    UNKNOWN = (None, "Unknown")

    @classmethod
    def _missing_(cls, _: Any) -> LitterBoxStatus:
        _LOGGER.error('Unknown status code "%s"', _)
        return cls.UNKNOWN

    @property
    def text(self) -> str | None:
        """Return the textual representation of a litter box's status."""
        return self._text

    @property
    def minimum_cycles_left(self) -> int:
        """Return the minimum number of cycles left based on a litter box's status."""
        return self._minimum_cycles_left

    @classmethod
    def get_drawer_full_statuses(
        cls,
        completely_full: bool = True,
        almost_full: bool = True,
        codes_only: bool = False,
    ) -> list[LitterBoxStatus | str]:
        """Return the statuses that represent that the waste drawer is full."""
        return [
            status.value if codes_only else status
            for status in (
                [cls.DRAWER_FULL, cls.STARTUP_DRAWER_FULL] if completely_full else []
            )
            + ([cls.DRAWER_FULL_1, cls.DRAWER_FULL_2] if almost_full else [])
        ]



"""Event handling class for pylitterbot."""

from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass, field
from typing import Any

EVENT_UPDATE = "update"


@dataclass
class Event:
    """Abstract event class properties and methods."""

    _listeners: dict[str, list[Callable]] = field(default_factory=dict)

    def emit(self, event_name: str, *args: Any, **kwargs: Any) -> None:
        """Run all callbacks for an event."""
        for listener in self._listeners.get(event_name, []):
            try:
                listener(*args, **kwargs)
            except:  # pragma: no cover # pylint: disable=bare-except # noqa: E722
                pass

    def on(  # pylint: disable=invalid-name
        self, event_name: str, callback: Callable
    ) -> Callable:
        """Register an event callback."""
        listeners: list = self._listeners.setdefault(event_name, [])
        listeners.append(callback)

        def unsubscribe() -> None:
            """Unsubscribe listeners."""
            if callback in listeners:
                listeners.remove(callback)

        return unsubscribe



"""Robot base class."""

from __future__ import annotations

import logging
from abc import abstractmethod
from collections.abc import Callable
from datetime import datetime
from typing import TYPE_CHECKING, Any, cast

from aiohttp import ClientWebSocketResponse
from deepdiff import DeepDiff

from ..event import EVENT_UPDATE, Event
from ..utils import to_timestamp, urljoin

if TYPE_CHECKING:
    from ..account import Account

_LOGGER = logging.getLogger(__name__)


class Robot(Event):
    """Robot base class."""

    _attr_model: str

    _data_id: str
    _data_name: str
    _data_serial: str
    _data_setup_date: str

    _path: str

    def __init__(self, data: dict, account: Account) -> None:
        """Initialize a robot."""
        super().__init__()
        self._data: dict = {}
        self._account = account

        self._is_loaded = False

        self._ws: ClientWebSocketResponse | None = None
        self._ws_subscription_id: str | None = None

        if data:
            if data.get(self._data_serial) is None:
                raise ValueError("Robot data must include a serial number")
            self._update_data(data)

    def __str__(self) -> str:
        """Return str(self)."""
        return f"Name: {self.name}, Model: {self.model}, Serial: {self.serial}, id: {self.id}"

    @property
    def id(self) -> str:  # pylint: disable=invalid-name
        """Return the id of the robot."""
        return str(self._data[self._data_id])

    @property
    @abstractmethod
    def is_online(self) -> bool:
        """Return `True` if the robot is online."""

    @property
    def model(self) -> str:
        """Return the robot model."""
        return self._attr_model

    @property
    def name(self) -> str:
        """Return the name of the robot, if any."""
        return cast(str, self._data.get(self._data_name, ""))

    @property
    @abstractmethod
    def night_light_mode_enabled(self) -> bool:
        """Return `True` if night light mode is enabled."""

    @property
    @abstractmethod
    def panel_lock_enabled(self) -> bool:
        """Return `True` if the buttons on the robot are disabled."""

    @property
    @abstractmethod
    def power_status(self) -> str:
        """Return the power type.

        `AC` = normal/mains
        `DC` = battery backup
        `NC` = unknown, not connected or off
        """

    @property
    def serial(self) -> str:
        """Return the serial of the robot, if any."""
        return cast(str, self._data.get(self._data_serial, ""))

    @property
    def setup_date(self) -> datetime | None:
        """Return the datetime the robot was onboarded, if any."""
        return to_timestamp(self._data.get(self._data_setup_date))

    @abstractmethod
    async def refresh(self) -> None:
        """Refresh the robot data from the API."""

    @abstractmethod
    async def set_name(self, name: str) -> bool:
        """Set the name."""

    @abstractmethod
    async def set_night_light(self, value: bool) -> bool:
        """Turn the night light mode on or off."""

    @abstractmethod
    async def set_panel_lockout(self, value: bool) -> bool:
        """Turn the panel lock on or off."""

    @abstractmethod
    async def send_subscribe_request(self, send_stop: bool = False) -> None:
        """Send a subscribe request and, optionally, unsubscribe from a previous subscription."""

    async def subscribe(self) -> None:
        """Open a web socket connection to receive updates."""
        try:
            self._ws = await self._account.ws_connect(self)
            await self.send_subscribe_request()
            _LOGGER.debug("%s subscribed to updates", self.name)
        except Exception as ex:  # pylint: disable=broad-except
            _LOGGER.error(ex)

    async def send_unsubscribe_request(self) -> None:
        """Send an unsubscribe request."""
        if self._ws and self._ws_subscription_id:
            await self._ws.send_json({"id": self._ws_subscription_id, "type": "stop"})

    async def unsubscribe(self) -> None:
        """Unsubscribe from the web socket."""
        if self._ws is not None and not self._ws.closed:
            await self.send_unsubscribe_request()
            self._ws = None
            _LOGGER.debug("%s unsubscribed from updates", self.name)

    def _update_data(
        self,
        data: dict,
        partial: bool = False,
        callback: Callable[[], Any] | None = None,
    ) -> None:
        """Save the robot info from a data dictionary."""
        if self._is_loaded:
            if diff := DeepDiff(
                self._data,
                {**self._data, **data} if partial else data,
                ignore_order=True,
                report_repetition=True,
                verbose_level=2,
            ):
                _LOGGER.debug("%s updated: %s", self.name, diff)

        self._data.update(data)
        if callback:
            callback()
        self._is_loaded = True
        self.emit(EVENT_UPDATE)

    async def _get(
        self, subpath: str | None = None, **kwargs: Any
    ) -> dict | list[dict] | None:
        """Send a GET request to the Litter-Robot API."""
        return await self._account.session.get(urljoin(self._path, subpath), **kwargs)

    async def _patch(
        self, subpath: str | None = None, json: Any | None = None, **kwargs: Any
    ) -> dict | list[dict] | None:
        """Send a PATCH request to the Litter-Robot API."""
        return await self._account.session.patch(
            urljoin(self._path, subpath), json=json, **kwargs
        )

    async def _post(
        self, subpath: str | None = None, json: Any | None = None, **kwargs: Any
    ) -> dict | list[dict] | None:
        """Send a POST request to the Litter-Robot API."""
        return await self._account.session.post(
            urljoin(self._path, subpath), json=json, **kwargs
        )

    @staticmethod
    async def get_websocket_config(account: Account) -> dict[str, Any]:
        """Get wesocket config."""
        raise NotImplementedError()

    @staticmethod
    def parse_websocket_message(data: dict) -> dict | None:
        """Parse a wesocket message."""
        raise NotImplementedError()



"""Litter-Robot base class."""

from __future__ import annotations

import logging
from abc import abstractmethod
from collections.abc import Callable
from datetime import datetime, time
from typing import Any, cast

from ..activity import Activity, Insight
from ..enums import LitterBoxCommand, LitterBoxStatus
from ..utils import to_timestamp
from . import Robot

_LOGGER = logging.getLogger(__name__)


MINIMUM_CYCLES_LEFT_DEFAULT = 3


class LitterRobot(Robot):
    """Data and methods for interacting with a Litter-Robot self-cleaning litter box."""

    VALID_WAIT_TIMES = [3, 7, 15]

    _data_cycle_capacity = "cycleCapacity"
    _data_cycle_capacity_default = 30
    _data_cycle_count = "cycleCount"
    _data_drawer_full_cycles = "cyclesAfterDrawerFull"
    _data_id = "litterRobotId"
    _data_name = "litterRobotNickname"
    _data_power_status = "powerStatus"
    _data_serial = "litterRobotSerial"
    _data_setup_date = "setupDate"

    _command_clean = LitterBoxCommand.CLEAN
    _command_night_light_off = LitterBoxCommand.NIGHT_LIGHT_OFF
    _command_night_light_on = LitterBoxCommand.NIGHT_LIGHT_ON
    _command_panel_lock_off = LitterBoxCommand.LOCK_OFF
    _command_panel_lock_on = LitterBoxCommand.LOCK_ON
    _command_power_off = LitterBoxCommand.POWER_OFF
    _command_power_on = LitterBoxCommand.POWER_ON

    _minimum_cycles_left: int = MINIMUM_CYCLES_LEFT_DEFAULT
    _sleep_mode_start_time: datetime | None = None
    _sleep_mode_end_time: datetime | None = None

    @property
    @abstractmethod
    def clean_cycle_wait_time_minutes(self) -> int:
        """Return the number of minutes after a cat uses the Litter-Robot to begin an automatic clean cycle."""

    @property
    def cycle_capacity(self) -> int:
        """Return the total anticpated number of clean cycles that can be performed before the waste drawer is full."""
        return int(
            self._data.get(self._data_cycle_capacity, self._data_cycle_capacity_default)
        )

    @property
    def cycle_count(self) -> int:
        """Return the cycle count since the last time the waste drawer was reset."""
        return int(self._data.get(self._data_cycle_count, 0))

    @property
    def cycles_after_drawer_full(self) -> int:
        """Return the cycles after the drawer is full."""
        return int(self._data.get(self._data_drawer_full_cycles, 0))

    @property
    @abstractmethod
    def is_drawer_full_indicator_triggered(self) -> bool:
        """Return `True` if the drawer full indicator has been triggered."""

    @property
    def is_onboarded(self) -> bool:
        """Return `True` if the Litter-Robot is onboarded."""
        return self._data.get("isOnboarded") is True

    @property
    @abstractmethod
    def is_sleeping(self) -> bool:
        """Return `True` if the Litter-Robot is currently "sleeping" and won't automatically perform a clean cycle."""

    @property
    @abstractmethod
    def is_waste_drawer_full(self) -> bool:
        """Return `True` if the Litter-Robot is reporting that the waste drawer is full."""

    @property
    def last_seen(self) -> datetime | None:
        """Return the datetime the Litter-Robot last reported, if any."""
        return to_timestamp(self._data.get("lastSeen"))

    @property
    def power_status(self) -> str:
        """Return the power status.

        `AC` = normal/mains
        `DC` = battery backup
        `NC` = unknown, not connected or off
        """
        return cast(str, self._data.get(self._data_power_status, "NC"))

    @property
    @abstractmethod
    def sleep_mode_enabled(self) -> bool:
        """Return `True` if sleep mode is enabled."""

    @property
    def sleep_mode_start_time(self) -> datetime | None:
        """Return the sleep mode start time, if any."""
        return self._sleep_mode_start_time

    @property
    def sleep_mode_end_time(self) -> datetime | None:
        """Return the sleep mode end time, if any."""
        return self._sleep_mode_end_time

    @property
    @abstractmethod
    def status(self) -> LitterBoxStatus:
        """Return the status of the Litter-Robot."""

    @property
    @abstractmethod
    def status_code(self) -> str | None:
        """Return the status code of the Litter-Robot."""

    @property
    def status_text(self) -> str | None:
        """Return the status text of the Litter-Robot."""
        return self.status.text

    @property
    @abstractmethod
    def waste_drawer_level(self) -> float:
        """Return the approximate waste drawer level."""

    def _update_data(
        self,
        data: dict,
        partial: bool = False,
        callback: Callable[[], Any] | None = None,
    ) -> None:
        """Save the Litter-Robot info from a data dictionary."""

        def _callback() -> None:
            self._parse_sleep_info()
            self._update_minimum_cycles_left()

        super()._update_data(data, partial, _callback)

    @abstractmethod
    def _parse_sleep_info(self) -> None:
        """Parse the sleep info of a Litter-Robot."""

    def _update_minimum_cycles_left(self) -> None:
        """Update the minimum cycles left."""
        if (
            self.status == LitterBoxStatus.READY
            or self._minimum_cycles_left > self.status.minimum_cycles_left
        ):
            self._minimum_cycles_left = self.status.minimum_cycles_left

    @abstractmethod
    async def _dispatch_command(self, command: str, **kwargs: Any) -> bool:
        """Send a command to the Litter-Robot."""

    async def start_cleaning(self) -> bool:
        """Start a cleaning cycle."""
        return await self._dispatch_command(self._command_clean)

    async def reset_settings(self) -> bool:  # pragma: no cover
        """Set the Litter-Robot back to default settings."""
        raise NotImplementedError()

    async def set_night_light(self, value: bool) -> bool:
        """Turn the night light mode on or off."""
        return await self._dispatch_command(
            self._command_night_light_on if value else self._command_night_light_off
        )

    async def set_panel_lockout(self, value: bool) -> bool:
        """Turn the panel lock on or off."""
        return await self._dispatch_command(
            self._command_panel_lock_on if value else self._command_panel_lock_off
        )

    async def set_power_status(self, value: bool) -> bool:
        """Turn the Litter-Robot on or off."""
        return await self._dispatch_command(
            self._command_power_on if value else self._command_power_off
        )

    async def set_sleep_mode(
        self, value: bool, sleep_time: time | None = None
    ) -> bool:  # pragma: no cover
        """Set the sleep mode on the Litter-Robot."""
        raise NotImplementedError()

    @abstractmethod
    async def set_wait_time(self, wait_time: int) -> bool:
        """Set the wait time on the Litter-Robot."""

    @abstractmethod
    async def get_activity_history(self, limit: int = 100) -> list[Activity]:
        """Return the activity history."""

    @abstractmethod
    async def get_insight(
        self, days: int = 30, timezone_offset: int | None = None
    ) -> Insight:
        """Return the insight data."""



"""Account access and data handling for Litter-Robot endpoint."""

from __future__ import annotations

import asyncio
import logging
from collections.abc import Callable
from typing import TypeVar, cast

from aiohttp import (
    ClientConnectorError,
    ClientResponseError,
    ClientSession,
    ClientWebSocketResponse,
)
from botocore.exceptions import ClientError

from .event import EVENT_UPDATE
from .exceptions import LitterRobotException, LitterRobotLoginException
from .pet import Pet
from .robot import Robot
from .robot.feederrobot import FEEDER_ENDPOINT, FEEDER_ROBOT_MODEL, FeederRobot
from .robot.litterrobot3 import DEFAULT_ENDPOINT, DEFAULT_ENDPOINT_KEY, LitterRobot3
from .robot.litterrobot4 import LITTER_ROBOT_4_MODEL, LR4_ENDPOINT, LitterRobot4
from .session import LitterRobotSession
from .utils import decode, urljoin
from .ws_monitor import WebSocketMonitor

_LOGGER = logging.getLogger(__name__)
_RobotT = TypeVar("_RobotT", bound=Robot)


class Account:
    """Class with data and methods for interacting with a user's Litter-Robots."""

    def __init__(
        self,
        token: dict | None = None,
        websession: ClientSession | None = None,
        token_update_callback: Callable[[dict | None], None] | None = None,
    ) -> None:
        """Initialize the account data."""
        self._session = LitterRobotSession(token=token, websession=websession)
        self._session._custom_args[DEFAULT_ENDPOINT] = {
            "headers": {"x-api-key": decode(DEFAULT_ENDPOINT_KEY)}
        }
        self._user_id = self._session.get_user_id() if token else None
        self._user: dict = {}
        self._robots: list[Robot] = []
        self._pets: list[Pet] = []
        self._monitors: dict[type[Robot], WebSocketMonitor] = {}

        if token_update_callback:
            self._session.on(
                EVENT_UPDATE,
                lambda session=self._session: token_update_callback(session.tokens),
            )

    @property
    def user_id(self) -> str | None:
        """Return the logged in user's id."""
        if not self._user_id and self.session:
            self._user_id = self.session.get_user_id()
        return self._user_id

    @property
    def robots(self) -> list[Robot]:
        """Return the set of robots for the logged in account."""
        return self._robots

    @property
    def pets(self) -> list[Pet]:
        """Return the set of pets for the logged in account."""
        return self._pets

    @property
    def session(self) -> LitterRobotSession:
        """Return the associated session on the account."""
        return self._session

    def get_robot(self, robot_id: str | int | None) -> Robot | None:
        """If found, return the robot with the specified id."""
        return next(
            (robot for robot in self._robots if robot.id == str(robot_id)),
            None,
        )

    def get_robots(self, robot_class: type[_RobotT]) -> list[_RobotT]:
        """If found, return the specified class of robots."""
        return [robot for robot in self._robots if isinstance(robot, robot_class)]

    def get_pet(self, pet_id: str) -> Pet | None:
        """If found, return the pet with the specified id."""
        return next(
            (pet for pet in self._pets if pet.id == pet_id),
            None,
        )

    async def connect(
        self,
        username: str | None = None,
        password: str | None = None,
        load_robots: bool = False,
        subscribe_for_updates: bool = False,
        load_pets: bool = False,
    ) -> None:
        """Connect to the Litter-Robot API."""
        try:
            if not self.session.is_token_valid():
                if self.session.has_refresh_token():
                    await self.session.refresh_tokens()
                elif username and password:
                    await self.session.login(username=username, password=password)
                else:
                    raise LitterRobotLoginException(
                        "Username and password are required to login to Litter-Robot."
                    )

            if load_robots:
                await self.load_robots(subscribe_for_updates)

            if load_pets:
                await self.load_pets()

        except ClientError as err:
            _LOGGER.error(err)
            raise LitterRobotLoginException(
                f"Unable to login to Litter-Robot: {err.response['message']}"
            ) from err
        except ClientResponseError as ex:
            _LOGGER.error(ex)
            if ex.status == 401:
                raise LitterRobotLoginException(
                    "Unable to login to Litter-Robot with the supplied credentials."
                ) from ex
            raise LitterRobotException("Unable to login to Litter-Robot.") from ex
        except ClientConnectorError as ex:
            _LOGGER.error(ex)
            raise LitterRobotException("Unable to reach the Litter-Robot api.") from ex

    async def disconnect(self) -> None:
        """Close the underlying session."""
        await asyncio.gather(*(robot.unsubscribe() for robot in self.robots))
        await asyncio.gather(*(monitor.close() for monitor in self._monitors.values()))
        await self.session.close()

    async def refresh_user(self) -> None:
        """Refresh the logged in user's info."""
        data = cast(
            dict,
            await self.session.get(urljoin(DEFAULT_ENDPOINT, f"users/{self.user_id}")),
        )
        self._user.update(data.get("user", {}))

    async def load_pets(self) -> None:
        """Get information about the pets connected to the account."""
        assert self.user_id
        pets = await Pet.fetch_pets_for_user(self._session, self.user_id)
        if not self._pets:
            self._pets = pets
        else:
            for pet in pets:
                if existing_pet := self.get_pet(pet.id):
                    existing_pet._update_data(pet._data)
                else:
                    self._pets.append(pet)

    async def load_robots(self, subscribe_for_updates: bool = False) -> None:
        """Get information about robots connected to the account."""
        robots: list[Robot] = []
        try:
            all_robots = [
                self.session.get(
                    urljoin(DEFAULT_ENDPOINT, f"users/{self.user_id}/robots")
                ),
                self.session.post(
                    LR4_ENDPOINT,
                    json={
                        "query": f"""
                            query GetLR4($userId: String!) {{
                                getLitterRobot4ByUser(userId: $userId) {LITTER_ROBOT_4_MODEL}
                            }}
                        """,
                        "variables": {"userId": self.user_id},
                    },
                ),
                self.session.post(
                    FEEDER_ENDPOINT,
                    json={
                        "query": f"""
                            query GetFeeders {{
                                feeder_unit {FEEDER_ROBOT_MODEL}
                            }}
                        """
                    },
                ),
            ]
            resp = await asyncio.gather(*all_robots)

            async def update_or_create_robot(
                robot_cls: type[Robot], data: dict
            ) -> None:
                # pylint: disable=protected-access
                if data.get(robot_cls._data_serial) is None:
                    _LOGGER.info(
                        "skipping robot without serial number (id=%s, name=%s)",
                        data.get(robot_cls._data_id),
                        data.get(robot_cls._data_name),
                    )
                    return
                if robot := self.get_robot(data.get(robot_cls._data_id)):
                    robot._update_data(data)
                else:
                    robot = robot_cls(data=data, account=self)
                    if subscribe_for_updates:
                        await robot.subscribe()
                robots.append(robot)

            for robot_data in resp[0]:  # type: ignore
                await update_or_create_robot(LitterRobot3, robot_data)
            for robot_data in resp[1].get("data").get("getLitterRobot4ByUser") or []:  # type: ignore
                await update_or_create_robot(LitterRobot4, robot_data)
            for robot_data in resp[2].get("data", {}).get("feeder_unit") or []:  # type: ignore
                await update_or_create_robot(FeederRobot, robot_data)

            self._robots = robots
        except (LitterRobotException, ClientResponseError, ClientConnectorError) as ex:
            _LOGGER.error("Unable to retrieve your robots: %s", ex)

    async def refresh_robots(self) -> None:
        """Refresh known robots."""
        try:
            await asyncio.gather(*(robot.refresh() for robot in self.robots))
        except (LitterRobotException, ClientResponseError, ClientConnectorError) as ex:
            _LOGGER.error("Unable to refresh your robots: %s", ex)

    async def get_bearer_authorization(self) -> str | None:
        """Return the authorization token."""
        if not self.session.is_token_valid():
            await self.session.refresh_tokens()
        return await self.session.get_bearer_authorization()

    async def ws_connect(self, robot: Robot) -> ClientWebSocketResponse:
        """Initiate a websocket connection for a robot."""
        robot_class = type(robot)
        ws_monitor = self._monitors.setdefault(
            robot_class, WebSocketMonitor(self, robot_class)
        )
        if ws_monitor.websocket is None or ws_monitor.websocket.closed:
            await ws_monitor.new_connection(True)
        if ws_monitor.monitor is None or ws_monitor.monitor.done():
            await ws_monitor.start_monitor()
        return ws_monitor.websocket



"""Pet profiles."""

from __future__ import annotations

import logging
from dataclasses import dataclass
from datetime import date, datetime
from enum import Enum, unique
from typing import Any, cast

from deepdiff import DeepDiff

from .event import EVENT_UPDATE, Event
from .exceptions import InvalidCommandException
from .session import Session
from .utils import to_timestamp

_LOGGER = logging.getLogger(__name__)

PET_PROFILE_ENDPOINT = "https://pet-profile.iothings.site/graphql/"

PET_MODEL = """
{
    petId
    userId
    createdAt
    name
    type
    gender
    weight
    weightLastUpdated
    lastWeightReading
    breeds
    age
    birthday
    adoptionDate
    s3ImageURL
    diet
    isFixed
    environmentType
    healthConcerns
    isHealthy
    isActive
    whiskerProducts
    petTagId
    weightIdFeatureEnabled
    weightHistory {
        weight
        timestamp
    }
    weightHistoryErrorType
}
"""


@unique
class PetDiet(Enum):
    """Pet diet."""

    WET = "WET_FOOD"
    DRY = "DRY_FOOD"
    BOTH = "BOTH"

    def __str__(self) -> str:
        """Return str(self)."""
        return "wet/dry" if self == PetDiet.BOTH else self.name.lower()


@unique
class PetEnvironment(Enum):
    """Pet environment."""

    INDOOR = "INDOOR"
    OUTDOOR = "OUTDOOR"
    BOTH = "BOTH"

    def __str__(self) -> str:
        """Return str(self)."""
        return "indoor/outdoor" if self == PetEnvironment.BOTH else self.name.lower()


@unique
class PetGender(Enum):
    """Pet gender."""

    FEMALE = "FEMALE"
    MALE = "MALE"

    def __str__(self) -> str:
        """Return str(self)."""
        return self.name.lower()


@unique
class PetType(Enum):
    """Pet type."""

    CAT = "CAT"
    DOG = "DOG"

    def __str__(self) -> str:
        """Return str(self)."""
        return self.name.lower()


@dataclass(frozen=True)
class WeightMeasurement:
    """Weight measurement."""

    timestamp: datetime
    weight: float

    def __str__(self) -> str:
        """Return self(str)."""
        return f"{self.timestamp.isoformat()}: {self.weight} lbs"


def parse_weight_history(
    weight_data: list[dict[str, Any]] | None,
) -> list[WeightMeasurement]:
    """Parse weight history from a list."""
    if not weight_data:
        return []
    return [
        WeightMeasurement(
            cast(datetime, to_timestamp(entry["timestamp"])), entry["weight"]
        )
        for entry in weight_data
        if entry["timestamp"]
    ]


class Pet(Event):
    """Pet profile."""

    def __init__(self, data: dict, session: Session) -> None:
        """Initialize a pet profile."""
        super().__init__()
        self._data: dict = data
        self._session: Session = session
        self._weight_history: list[WeightMeasurement] = []
        if weight_data := data.get("weightHistory"):
            self._weight_history = parse_weight_history(weight_data)

    def __str__(self) -> str:
        """Return str(self)."""
        return f"Name: {self.name}, Gender: {self.gender}, Type: {self.pet_type}, Breed: {self.breeds}, id: {self.id}"

    @property
    def id(self) -> str:
        """Return the id of the pet."""
        return cast(str, self._data.get("petId"))

    @property
    def name(self) -> str:
        """Return the name of the pet."""
        return cast(str, self._data.get("name"))

    @property
    def pet_type(self) -> PetType | None:
        """Return the type of pet."""
        if (pet_type := self._data.get("type")) not in PetType.__members__:
            _LOGGER.error('Unknown pet type "%s"', pet_type)
            return None
        return PetType(pet_type)

    @property
    def gender(self) -> PetGender | None:
        """Return the gender."""
        if (gender := self._data.get("gender")) not in PetGender.__members__:
            _LOGGER.error('Unknown gender "%s"', gender)
            return None
        return PetGender(gender)

    @property
    def estimated_weight(self) -> float:
        """Return the estimated weight in pounds (lbs)."""
        return cast(float, self._data.get("weight"))

    @property
    def last_weight_reading(self) -> float | None:
        """Return the last weight reading in pounds (lbs), if any."""
        return self._data.get("lastWeightReading")

    @property
    def weight(self) -> float:
        """Return the weight in pounds (lbs)."""
        return self.last_weight_reading or self.estimated_weight

    @property
    def breeds(self) -> list[str] | None:
        """Return the breeds, if known."""
        return self._data.get("breeds")

    @property
    def age(self) -> int | None:
        """Return the age, if known."""
        return self._data.get("age")

    @property
    def birthday(self) -> date | None:
        """Return the birthday, if known."""
        birthday_str = self._data.get("birthday")
        return datetime.fromisoformat(birthday_str).date() if birthday_str else None

    @property
    def adoption_date(self) -> date | None:
        """Return the adoption date, if known."""
        adoption_date_str = self._data.get("adoptionDate")
        return (
            datetime.fromisoformat(adoption_date_str).date()
            if adoption_date_str
            else None
        )

    @property
    def diet(self) -> PetDiet | None:
        """Return the diet, if any."""
        if (diet := self._data.get("diet")) not in PetDiet.__members__:
            _LOGGER.error('Unknown diet "%s"', diet)
            return None
        return PetDiet(diet)

    @property
    def environment_type(self) -> PetEnvironment | None:
        """Return the environment type, if any."""
        if (
            environment := self._data.get("environmentType")
        ) not in PetEnvironment.__members__:
            _LOGGER.error('Unknown environment type "%s"', environment)
            return None
        return PetEnvironment(environment)

    @property
    def health_concerns(self) -> list[str] | None:
        """Return a list of health concerns, if any."""
        return self._data.get("healthConcerns")

    @property
    def image_url(self) -> str | None:
        """Return image url, if any."""
        return self._data.get("s3ImageURL")

    @property
    def is_active(self) -> bool | None:
        """Return if the pet profile is active."""
        return cast(bool, self._data.get("isActive", False))

    @property
    def is_fixed(self) -> bool | None:
        """Return `True` if the pet is fixed."""
        return cast(bool, self._data.get("isFixed", False))

    @property
    def is_healthy(self) -> bool | None:
        """Return `True` if the pet is healthy."""
        return cast(bool, self._data.get("isHealthy", False))

    @property
    def pet_tag_id(self) -> str | None:
        """Return the pet tag id, if any."""
        return self._data.get("petTagID")

    @property
    def weight_id_feature_enabled(self) -> bool:
        """Return `True` if the weight id feature is enabled."""
        return cast(bool, self._data.get("weightIdFeatureEnabled", False))

    @property
    def weight_history(self) -> list[WeightMeasurement]:
        """Return the weight history for the pet."""
        return self._weight_history

    def get_visits_since(self, start: datetime) -> int:
        """Return the number of visits (recorded via weight history) since the given datetime."""
        return sum(entry.timestamp >= start for entry in self.weight_history)

    def _update_data(self, data: dict, partial: bool = False) -> None:
        """Save the pet info from a data dictionary."""
        if diff := DeepDiff(
            self._data,
            {**self._data, **data} if partial else data,
            ignore_order=True,
            report_repetition=True,
            verbose_level=2,
        ):
            _LOGGER.debug("%s updated: %s", self.name, diff)

        if weight_data := data.get("weightHistory"):
            self._weight_history = parse_weight_history(weight_data)
        self._data.update(data)
        self.emit(EVENT_UPDATE)

    async def fetch_weight_history(self, limit: int = 50) -> list[WeightMeasurement]:
        """Fetch a pet's weight history."""
        weight_data = await self.query_weight_history(self._session, self.id, limit)
        self._weight_history = parse_weight_history(weight_data)
        return self._weight_history

    async def refresh(self) -> None:
        """Refresh the data for the pet."""
        data = await self.query_by_id(self._session, self.id)
        self._update_data(data)

    @classmethod
    async def fetch_pets_for_user(cls, session: Session, user_id: str) -> list[Pet]:
        """Fetch pets for a user."""
        pets_data = await cls.query_by_user(session, user_id)
        return [cls(pet_data, session) for pet_data in pets_data]

    @classmethod
    async def fetch_pet_by_id(cls, session: Session, pet_id: str) -> Pet:
        """Fetch a pet by id."""
        pet_data = await cls.query_by_id(session, pet_id)
        pet = cls(pet_data, session)
        return pet

    @staticmethod
    async def query_by_user(session: Session, user_id: str) -> list[dict]:
        """Query pets for a user."""
        query = f"""
            query GetPetsByUser($userId: String!) {{
                getPetsByUser(userId: $userId ) {PET_MODEL}
            }}
        """
        vars = {"userId": user_id}

        res = cast(dict, await Pet.query_graphql_api(session, query, vars))
        return cast(list[dict], res.get("data", {}).get("getPetsByUser", []))

    @staticmethod
    async def query_by_id(session: Session, pet_id: str) -> dict:
        """Query a pet by id."""
        query = f"""
            query GetPetByPetId($petId: String!) {{
                getPetByPetId(petId: $petId ) {PET_MODEL}
            }}
        """
        vars = {"petId": pet_id}

        res = cast(dict, await Pet.query_graphql_api(session, query, vars))
        return cast(dict, res.get("data", {}).get("getPetByPetId", {}))

    @staticmethod
    async def query_weight_history(
        session: Session, pet_id: str, limit: int = 50
    ) -> list[dict]:
        """Query a pet's weight history."""
        if limit < 1:
            raise InvalidCommandException(
                f"Invalid range for parameter limit, value: {limit}, valid range: 1-inf"
            )

        query = """
            query GetWeightHistoryByPetId($petId: String!, $limit: Int) {
                getWeightHistoryByPetId(petId: $petId, limit: $limit) {
                    weight
                    timestamp
                }
            }
            """
        vars = {"petId": pet_id, "limit": limit}

        res = cast(dict, await Pet.query_graphql_api(session, query, vars))
        return cast(list[dict], res.get("data", {}).get("getWeightHistoryByPetId", []))

    @staticmethod
    async def query_graphql_api(
        session: Session,
        query: str,
        variables: dict | None = None,
        endpoint: str = PET_PROFILE_ENDPOINT,
    ) -> dict | list[dict] | None:
        """Query GraphQL API."""
        return await session.post(
            endpoint, json={"query": query, "variables": variables}
        )



"""Feeder-Robot."""

from __future__ import annotations

import logging
from copy import deepcopy
from datetime import datetime, time, timedelta
from typing import TYPE_CHECKING, Any, Callable, TypeVar, cast
from uuid import uuid4
from zoneinfo import ZoneInfo

from ..enums import FeederRobotCommand
from ..exceptions import InvalidCommandException
from ..utils import decode, to_timestamp, utcnow
from . import Robot
from .models import FEEDER_ROBOT_MODEL

if TYPE_CHECKING:
    from ..account import Account

_T = TypeVar("_T")

_LOGGER = logging.getLogger(__name__)

FEEDER_ENDPOINT = "https://cognito.hasura.iothings.site/v1/graphql"
COMMAND_ENDPOINT = (
    "https://42nk7qrhdg.execute-api.us-east-1.amazonaws.com/prod/command/feeder"
)
COMMAND_ENDPOINT_KEY = decode(
    "dzJ0UEZiamxQMTNHVW1iOGRNalVMNUIyWXlQVkQzcEo3RXk2Zno4dg=="
)

FOOD_LEVEL_MAP = {9: 100, 8: 70, 7: 60, 6: 50, 5: 40, 4: 30, 3: 20, 2: 10, 1: 5, 0: 0}
MEAL_INSERT_SIZE_CUPS_MAP = {0: 1 / 4, 1: 1 / 8}
MEAL_INSERT_SIZE_CUPS_REVERSE_MAP = {v: k for k, v in MEAL_INSERT_SIZE_CUPS_MAP.items()}

WEEKDAY_MAP = {
    "Mon": 0,
    "Tue": 1,
    "Wed": 2,
    "Thu": 3,
    "Fri": 4,
    "Sat": 5,
    "Sun": 6,
}


class FeederRobot(Robot):  # pylint: disable=abstract-method
    """Data and methods for interacting with a Feeder-Robot automatic pet feeder."""

    _attr_model = "Feeder-Robot"

    VALID_MEAL_INSERT_SIZES = list(MEAL_INSERT_SIZE_CUPS_MAP.values())

    _data_id = "id"
    _data_name = "name"
    _data_serial = "serial"
    _data_setup_date = "created_at"

    _last_updated_at: str | None = None
    _next_feeding: datetime | None = None

    def __init__(self, data: dict, account: Account) -> None:
        """Initialize a Feeder-Robot."""
        super().__init__(data, account)
        self._path = FEEDER_ENDPOINT

    def _state_info(self, key: str, default: _T | None = None) -> _T | Any:
        """Get an attribute from the data.state.info section."""
        return cast(_T, self._data["state"]["info"].get(key, default))

    @property
    def firmware(self) -> str:
        """Return the firmware version."""
        return str(self._state_info("fwVersion"))

    @property
    def food_level(self) -> int:
        """Return the food level."""
        return FOOD_LEVEL_MAP.get(self._state_info("level"), 0)

    @property
    def gravity_mode_enabled(self) -> bool:
        """Return `True` if gravity mode is enabled."""
        return bool(self._state_info("gravity"))

    @property
    def is_online(self) -> bool:
        """Return `True` if the robot is online."""
        return bool(self._state_info("online"))

    @property
    def last_feeding(self) -> dict[str, Any] | None:
        """Get the last feeding meal or snack dispensed."""
        meal = self.last_meal
        if (snack := self.last_snack) is None or (
            meal is not None and meal["timestamp"] > snack["timestamp"]
        ):
            return meal
        return snack

    @property
    def last_meal(self) -> dict[str, Any] | None:
        """Get the last meal dispensed."""
        if not (meals := self._data.get("feeding_meal")):
            return None
        return {
            "timestamp": to_timestamp(meals[0]["timestamp"]),
            "amount": meals[0]["amount"] * meals[0]["meal_total_portions"],
            "name": meals[0]["meal_name"],
        }

    @property
    def last_snack(self) -> dict[str, Any] | None:
        """Get the last snack dispensed."""
        if not (snacks := self._data.get("feeding_snack")):
            return None
        return {
            "timestamp": to_timestamp(snacks[0]["timestamp"]),
            "amount": snacks[0]["amount"],
            "name": "snack",
        }

    @property
    def meal_insert_size(self) -> float:
        """Return the meal insert size in cups."""
        meal_insert_size = self._state_info("mealInsertSize", 0)
        if not (cups := MEAL_INSERT_SIZE_CUPS_MAP.get(meal_insert_size, 0)):
            _LOGGER.error('Unknown meal insert size "%s"', meal_insert_size)
        return cups

    @property
    def next_feeding(self) -> datetime | None:
        """Return the next feeding, if any."""
        if self._next_feeding and self._next_feeding < utcnow():
            self._calculate_next_feeding()
        return self._next_feeding

    @property
    def night_light_mode_enabled(self) -> bool:
        """Return `True` if night light mode is enabled."""
        return bool(self._state_info("autoNightMode"))

    @property
    def panel_lock_enabled(self) -> bool:
        """Return `True` if the buttons on the robot are disabled."""
        return bool(self._state_info("panelLockout"))

    @property
    def power_status(self) -> str:
        """Return the power type.

        `AC` = normal/mains
        `DC` = battery backup
        `NC` = unknown, not connected or off
        """
        if bool(self._state_info("acPower")):
            return "AC"
        if bool(self._state_info("dcPower")):
            return "DC"
        return "NC"  # This *may* never happen

    @property
    def timezone(self) -> str:
        """Return the timezone."""
        return str(self._data.get("timezone"))

    @property
    def updated_at(self) -> str:
        """Return the updated at string."""
        return str(self._data["state"].get("updated_at"))

    def get_food_dispensed_since(self, start: datetime) -> float:
        """Return the amount of food (in cups) since the given datetime."""
        feedings: list[dict] = list(self._data.get("feeding_meal") or [])
        feedings += self._data.get("feeding_snack") or []
        amount: float = sum(
            feeding["amount"] * feeding.get("meal_total_portions", 1)
            for feeding in feedings
            if cast(datetime, to_timestamp(feeding["timestamp"])) >= start
        )
        return amount

    def _calculate_next_feeding(self) -> None:
        """Return the next scheduled feeding, if any."""
        if self.gravity_mode_enabled:
            self._next_feeding = None
            return

        schedule = self._data["state"].get("active_schedule")
        if not schedule or "meals" not in schedule:
            self._next_feeding = None
            return

        tz = ZoneInfo(self.timezone)
        now = datetime.now(tz)

        next_meal_time = None

        for meal in schedule["meals"]:
            if meal.get("paused"):
                continue

            # Skip meals with a skip date that is today or in the future
            skip = meal.get("skip")
            if skip and skip != "0000-01-01T00:00:00.000":
                skip_dt = datetime.fromisoformat(skip)
                if skip_dt.tzinfo is None:
                    skip_dt = skip_dt.replace(tzinfo=tz)
                if skip_dt.date() >= now.date():
                    continue

            meal_time = time(meal["hour"], meal["minute"])

            for day in meal["days"]:
                target_weekday = WEEKDAY_MAP[day]
                days_ahead = (target_weekday - now.weekday() + 7) % 7

                # If today, check if meal time has already passed
                if days_ahead == 0 and meal_time <= now.time():
                    days_ahead = 7

                feeding_datetime = datetime.combine(
                    now.date() + timedelta(days=days_ahead), meal_time
                ).replace(tzinfo=tz)

                if next_meal_time is None or feeding_datetime < next_meal_time:
                    next_meal_time = feeding_datetime

        self._next_feeding = next_meal_time

    def _update_data(
        self,
        data: dict,
        partial: bool = False,
        callback: Callable[[], Any] | None = None,
    ) -> None:
        """Save the Feeder-Robot info from a data dictionary."""

        def _callback() -> None:
            if callback:
                callback()
            if self._last_updated_at != self.updated_at:
                self._calculate_next_feeding()
                self._last_updated_at = self.updated_at

        super()._update_data(data, partial, _callback)

    async def _dispatch_command(self, command: str, value: bool) -> bool:
        """Send a command to the Feeder-Robot."""
        try:
            await self._post(
                COMMAND_ENDPOINT,
                json={
                    "command": command,
                    "id": str(uuid4()),
                    "serial": self.serial,
                    "value": 1 if value else 0,
                },
                headers={"x-api-key": COMMAND_ENDPOINT_KEY},
            )
            return True
        except InvalidCommandException as ex:
            _LOGGER.error(ex)
            return False

    async def give_snack(self) -> bool:
        """Dispense a snack."""
        return await self._dispatch_command(FeederRobotCommand.GIVE_SNACK, True)

    async def refresh(self) -> None:
        """Refresh the Feeder-Robot's data from the API."""
        data = await self._post(
            json={
                "query": f"""
                    query GetFeeder($id: Int!) {{
                        feeder_unit_by_pk(id: $id) {FEEDER_ROBOT_MODEL}
                    }}
                    """,
                "variables": {"id": self.id},
            },
        )
        self._update_data(cast(dict, data).get("data", {}).get("feeder_unit_by_pk", {}))

    async def set_meal_insert_size(self, meal_insert_size: float) -> bool:
        """Set the meal insert size."""
        if (value := MEAL_INSERT_SIZE_CUPS_REVERSE_MAP.get(meal_insert_size)) is None:
            raise InvalidCommandException(
                f"Only meal insert sizes of {list(MEAL_INSERT_SIZE_CUPS_REVERSE_MAP)} are supported."
            )

        data = await self._post(
            json={
                "query": """
                    mutation UpdateFeederState($id: Int!, $state: jsonb) {
                        update_feeder_unit_state_by_pk(pk_columns: {id: $id}, _append: {info: $state}) {
                            info
                            updated_at
                        }
                    }
                """,
                "variables": {
                    "id": self._data["state"]["id"],
                    "state": {
                        "mealInsertSize": value,
                        "historyInvalidationDate": utcnow().strftime(
                            "%Y-%m-%dT%H:%M:%S.%fZ"
                        ),
                    },
                },
            }
        )
        data = cast(dict, data)
        self._update_data(
            {
                **self._data,
                "state": {
                    **self._data["state"],
                    **data.get("data", {}).get("update_feeder_unit_state_by_pk", {}),
                },
            }
        )
        return self.meal_insert_size == meal_insert_size

    async def set_name(self, name: str) -> bool:
        """Set the name."""
        data = await self._post(
            json={
                "query": """
                    mutation UpdateFeeder($id: Int!, $name: String!) {
                        update_feeder_unit_by_pk(pk_columns: {id: $id}, _set: {name: $name}) {
                            name
                        }
                    }
                """,
                "variables": {"id": self.id, "name": name},
            }
        )
        data = cast(dict, data)
        self._update_data(
            {**self._data, **data.get("data", {}).get("update_feeder_unit_by_pk", {})}
        )
        return self.name == name

    async def set_gravity_mode(self, value: bool) -> bool:
        """Turn the gravity mode on or off."""
        if await self._dispatch_command(FeederRobotCommand.SET_GRAVITY_MODE, value):
            data = deepcopy(self._data)
            data["state"]["info"]["gravity"] = value
            data["state"]["updated_at"] = utcnow().isoformat()
            self._update_data(data)
        return self.gravity_mode_enabled == value

    async def set_night_light(self, value: bool) -> bool:
        """Turn the night light mode on or off."""
        if await self._dispatch_command(FeederRobotCommand.SET_AUTO_NIGHT_MODE, value):
            data = deepcopy(self._data)
            data["state"]["info"]["autoNightMode"] = value
            self._update_data(data)
        return self.night_light_mode_enabled == value

    async def set_panel_lockout(self, value: bool) -> bool:
        """Turn the panel lock on or off."""
        if await self._dispatch_command(FeederRobotCommand.SET_PANEL_LOCKOUT, value):
            data = deepcopy(self._data)
            data["state"]["info"]["panelLockout"] = value
            self._update_data(data)
        return self.panel_lock_enabled == value

    async def send_subscribe_request(self, send_stop: bool = False) -> None:
        """Send a subscribe request and, optionally, unsubscribe from a previous subscription."""
        if not self._ws:
            return
        if send_stop:
            await self.send_unsubscribe_request()
        self._ws_subscription_id = str(uuid4())

        await self._ws.send_json(
            {
                "type": "connection_init",
                "payload": {
                    "headers": {
                        "Authorization": await self._account.get_bearer_authorization()
                    }
                },
            }
        )
        await self._ws.send_json(
            {
                "type": "start",
                "id": self._ws_subscription_id,
                "payload": {
                    "query": f"""
                            subscription GetFeeder($id: Int!) {{
                                feeder_unit_by_pk(id: $id) {FEEDER_ROBOT_MODEL}
                            }}
                        """,
                    "variables": {"id": self.id},
                },
            }
        )

    @staticmethod
    async def get_websocket_config(account: Account) -> dict[str, Any]:
        """Get wesocket config."""
        return {
            "url": FEEDER_ENDPOINT,
            "params": None,
            "headers": {"sec-websocket-protocol": "graphql-ws"},
        }

    @staticmethod
    def parse_websocket_message(data: dict) -> dict | None:
        """Parse a wesocket message."""
        if (data_type := data["type"]) == "data":
            return cast(dict, data["payload"]["data"]["feeder_unit_by_pk"])
        if data_type == "error":
            _LOGGER.error(data)
        elif data_type not in ("connection_ack", "ka", "complete"):
            _LOGGER.debug(data)
        return None



"""Session handling for litter-robot endpoint."""

from __future__ import annotations

import asyncio
import logging
from abc import ABC, abstractmethod
from asyncio import Lock
from functools import partial
from types import TracebackType
from typing import Any, Final, TypeVar, cast

import jwt
from aiohttp import ClientSession
from botocore.exceptions import ClientError, ParamValidationError
from pycognito import Cognito

from .event import EVENT_UPDATE, Event
from .exceptions import InvalidCommandException
from .utils import decode, redact, utcnow

T = TypeVar("T", bound="Session")

_LOGGER = logging.getLogger(__name__)


class Session(Event, ABC):
    """Abstract session class."""

    _lock = Lock()

    def __init__(self, websession: ClientSession | None = None) -> None:
        """Initialize the session."""
        super().__init__()
        self._websession_provided = websession is not None
        self._websession = websession

    @property
    def websession(self) -> ClientSession:
        """Get websession."""
        if self._websession is None:
            self._websession = ClientSession()
        return self._websession

    @property
    @abstractmethod
    def tokens(self) -> dict[str, str] | None:
        """Return the tokens."""

    async def close(self) -> None:
        """Close the session."""
        if not self._websession_provided and self.websession is not None:
            await self.websession.close()

    async def get(self, path: str, **kwargs: Any) -> dict | list[dict] | None:
        """Send a GET request to the specified path."""
        return await self.request("GET", path, **kwargs)

    async def post(self, path: str, **kwargs: Any) -> dict | list[dict] | None:
        """Send a POST request to the specified path."""
        return await self.request("POST", path, **kwargs)

    async def patch(self, path: str, **kwargs: Any) -> dict | list[dict] | None:
        """Send a PATCH request to the specified path."""
        return await self.request("PATCH", path, **kwargs)

    @abstractmethod
    async def async_get_id_token(self, **kwargs: Any) -> str | None:
        """Return a valid access token."""

    @abstractmethod
    def is_token_valid(self) -> bool:
        """Return `True` if the token is stills valid."""

    async def refresh_tokens(self, ignore_unexpired: bool = False) -> None:
        """Refresh the access token."""
        if self.tokens is None:
            return None
        async with self._lock:
            if not ignore_unexpired and self.is_token_valid():
                return
            await self._refresh_tokens()
        self.emit(EVENT_UPDATE)

    @abstractmethod
    async def _refresh_tokens(self) -> None:
        """Actual implementation to refresh the tokens."""

    async def get_bearer_authorization(self) -> str | None:
        """Get the bearer authorization."""
        if (access_token := await self.async_get_id_token()) is None:
            return None
        return f"Bearer {access_token}"

    async def request(
        self, method: str, url: str, **kwargs: Any
    ) -> dict | list[dict] | None:
        """Make a request."""
        _LOGGER.debug("Making %s request to %s", method, url)

        if "headers" not in kwargs:
            kwargs["headers"] = {}

        if (authorization := await self.get_bearer_authorization()) is not None:
            kwargs["headers"]["authorization"] = authorization

        async with self.websession.request(method, url, **kwargs) as resp:
            if resp.status == 500:
                if (data := await resp.json()).get("type") == "InvalidCommandException":
                    raise InvalidCommandException(data.get("developerMessage", data))
                raise InvalidCommandException(data)
            if resp.status == 401:
                if authorization is not None:
                    _LOGGER.error(
                        "Now: %s, Expiration: %s, Difference: %s",
                        (now := utcnow().timestamp()),
                        (
                            expires := jwt.decode(
                                authorization.replace("Bearer ", ""),
                                options={"verify_signature": False},
                            )["exp"]
                        ),
                        expires - now,
                    )
                _LOGGER.error("Unauthorized")

            resp.raise_for_status()
            data = await resp.json()
            _LOGGER.debug(
                "Received %s response from %s: %s", resp.status, url, redact(data)
            )
            return data  # type: ignore

    async def __aenter__(self: T) -> T:
        """Async enter."""
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Async exit."""
        await self.close()


class LitterRobotSession(Session):
    """Class with methods for interacting with a Litter-Robot cloud session."""

    USER_POOL_ID: Final = "dXMtZWFzdC0xX3JqaE5uWlZBbQ=="
    CLIENT_ID: Final = "NDU1MnVqZXUzYWljOTBuZjhxbjUzbGV2bW4="

    _user: Cognito | None = None

    _username: str | None = None
    __access_token: str | None = None
    __id_token: str | None = None
    __refresh_token: str | None = None

    def __init__(
        self, token: dict | None = None, websession: ClientSession | None = None
    ) -> None:
        """Initialize the session."""
        super().__init__(websession=websession)

        if token:
            self.__access_token = token.get("access_token")
            self.__id_token = token.get("id_token")
            self.__refresh_token = token.get("refresh_token")
        self._custom_args: dict = {}

    @property
    def access_token(self) -> str | None:
        """Return the access token, if any."""
        return self._user.access_token if self._user else self.__access_token

    @property
    def id_token(self) -> str | None:
        """Return the id token, if any."""
        return self._user.id_token if self._user else self.__id_token

    @property
    def refresh_token(self) -> str | None:
        """Return the refresh token, if any."""
        return self._user.refresh_token if self._user else self.__refresh_token

    @property
    def tokens(self) -> dict[str, str] | None:
        """Return the tokens."""
        if None in (self.access_token, self.id_token):
            return None
        token = {
            "access_token": self.access_token,
            "id_token": self.id_token,
            "refresh_token": self.refresh_token,
        }
        return cast(dict[str, str], token)

    def generate_args(self, url: str, **kwargs: Any) -> dict[str, Any]:
        """Generate args."""
        for key, value in next(
            (value for key, value in self._custom_args.items() if url.startswith(key)),
            {},
        ).items():
            if (orig := kwargs.get(key)) is not None:
                value = {**value, **orig} if isinstance(value, dict) else orig
            kwargs[key] = value
        return kwargs

    def is_token_valid(self) -> bool:
        """Return `True` if the token is stills valid."""
        if self.tokens is None:
            return False
        try:
            jwt.decode(
                self.id_token,
                options={"verify_signature": False, "verify_exp": True},
                leeway=-30,
            )
        except jwt.ExpiredSignatureError:
            return False
        return True

    async def async_get_id_token(self, **kwargs: Any) -> str | None:
        """Return a valid id token."""
        if self.tokens is None or not self.is_token_valid():
            return None
        return self.id_token

    async def login(self, username: str, password: str) -> None:
        """Login to the Litter-Robot api and generate a new token."""
        self._username = username
        user = await self.get_user()
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, partial(user.authenticate, password=password))
        self.emit(EVENT_UPDATE)

    async def _refresh_tokens(self) -> None:
        """Refresh the access token."""
        user = await self.get_user()
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, user.renew_access_token)

    async def request(
        self, method: str, url: str, **kwargs: Any
    ) -> dict | list[dict] | None:
        """Make a request."""
        kwargs = self.generate_args(url, **kwargs)
        if not kwargs.pop("skip_auth", False) and not self.is_token_valid():
            await self.refresh_tokens()
        return await super().request(method, url, **kwargs)

    async def get_user(self) -> Cognito:
        """Return the Cognito user."""
        if self._user is None:
            loop = asyncio.get_running_loop()
            self._user = await loop.run_in_executor(
                None,
                partial(
                    Cognito,
                    decode(self.USER_POOL_ID),
                    decode(self.CLIENT_ID),
                    username=self._username,
                    access_token=self.__access_token,
                    id_token=self.__id_token,
                    refresh_token=self.__refresh_token,
                ),
            )
            assert self._user
            if self.__access_token and self.__id_token:
                try:
                    await loop.run_in_executor(None, self._user.check_token)
                    self._user.verify_tokens()
                except ClientError as err:
                    _LOGGER.error(err)
                    raise err
                except ParamValidationError:
                    # tokens are invalid
                    pass
        if self._username and not self._user.username:
            self._user.username = self._username
        return self._user

    def get_user_id(self) -> str | None:
        """Get the user id from the session."""
        if self.tokens is None:
            return None
        user_id = jwt.decode(
            self.id_token,
            options={"verify_signature": False, "verify_exp": False},
        )["mid"]
        return cast(str, user_id)

    def has_refresh_token(self) -> bool:
        """Return `True` if the session has a refresh token."""
        return self.tokens is not None and self.refresh_token is not None
